import builtins
import collections.abc as c_abc
import inspect
from typing import (Any, Callable, Dict, Hashable, Iterable, Mapping, overload, Set,
                    TypeVar, Union)

from .._internal.argspec import Arguments
from .._internal.default_container import get_default_container
from .._internal.wrapper import InjectedWrapper, Injection, InjectionBlueprint
from ..core import DependencyContainer

F = TypeVar('F', Callable, staticmethod, classmethod)

_BUILTINS_TYPES = {e for e in builtins.__dict__.values() if isinstance(e, type)}
DEPENDENCIES_TYPE = Union[
    Mapping[str, Hashable],  # {arg_name: dependency, ...}
    Iterable[Hashable],  # (dependency for arg 1, ...)
    Callable[[str], Hashable],  # arg_name -> dependency
    str  # str.format(arg_name=arg_name) -> dependency
]


@overload
def inject(func: F,  # noqa: E704  # pragma: no cover
           *,
           arguments: Arguments = None,
           dependencies: DEPENDENCIES_TYPE = None,
           use_names: Union[bool, Iterable[str]] = None,
           use_type_hints: Union[bool, Iterable[str]] = None,
           container: DependencyContainer = None
           ) -> F: ...


@overload
def inject(*,  # noqa: E704  # pragma: no cover
           arguments: Arguments = None,
           dependencies: DEPENDENCIES_TYPE = None,
           use_names: Union[bool, Iterable[str]] = None,
           use_type_hints: Union[bool, Iterable[str]] = None,
           container: DependencyContainer = None
           ) -> Callable[[F], F]: ...


def inject(func=None,
           *,
           arguments: Arguments = None,
           dependencies: DEPENDENCIES_TYPE = None,
           use_names: Union[bool, Iterable[str]] = None,
           use_type_hints: Union[bool, Iterable[str]] = None,
           container: DependencyContainer = None
           ):
    """
    Inject the dependencies into the function lazily, they are only retrieved
    upon execution. Can be used as a decorator.

    Dependency CAN NOT be:

    - part of the builtins
    - part of typing
    - None

    Args:
        func: Callable to be wrapped.
        arguments: Arguments of the function can directly be specified if they
            have already been built.
        dependencies: Can be either a mapping of arguments name to their
            dependency, an iterable of dependencies or a function which returns
            the dependency given the arguments name. If an iterable is specified,
            the position of the arguments is used to determine their respective
            dependency. An argument may be skipped by using :code:`None` as a
            placeholder. The first argument is always ignored for methods (self)
            and class methods (cls).Type hints are overridden. Defaults to :code:`None`.
        use_names: Whether or not the arguments' name should be used as their
            respective dependency. An iterable of argument names may also be
            supplied to restrict this to those. Defaults to :code:`False`.
        use_type_hints: Whether or not the type hints (annotations) should be
            used as the arguments dependency. An iterable of argument names may
            also be specified to restrict this to those. Any type hints from
            the builtins (str, int...) or the typing (:py:class:`~typing.Optional`,
            ...) are ignored. Defaults to :code:`True`.
        container: :py:class:`~.core.container.DependencyContainer` from which
            the dependencies should be retrieved. Defaults to the global
            core if it is defined.

    Returns:
        The decorator to be applied or the injected function if the
        argument :code:`func` was supplied.

    """

    def _inject(wrapped):
        if inspect.isclass(wrapped):
            # @inject on a class would not return a class which is
            # counter-intuitive.
            raise TypeError("Classes cannot be wrapped with @inject. "
                            "Consider using @wire")

        nonlocal arguments
        # if the function has already its dependencies injected, no need to do
        # it twice.
        if isinstance(wrapped, InjectedWrapper):
            return wrapped

        if arguments is None:
            arguments = Arguments.from_callable(wrapped)

        blueprint = _build_injection_blueprint(
            arguments=arguments,
            dependencies=dependencies,
            use_names=use_names,
            use_type_hints=use_type_hints
        )

        # If nothing can be injected, just return the existing function without
        # any overhead.
        if all(injection.dependency is None for injection in blueprint.injections):
            return wrapped

        return InjectedWrapper(container=container or get_default_container(),
                               blueprint=blueprint,
                               wrapped=wrapped)

    return func and _inject(func) or _inject


def _build_injection_blueprint(arguments: Arguments,
                               dependencies: DEPENDENCIES_TYPE = None,
                               use_names: Union[bool, Iterable[str]] = None,
                               use_type_hints: Union[bool, Iterable[str]] = None,
                               ) -> InjectionBlueprint:
    """
    Construct a InjectionBlueprint with all the necessary information about
    the arguments for dependency injection. Storing it avoids significant
    execution overhead.

    Used by inject()
    """
    use_names = use_names if use_names is not None else False
    use_type_hints = use_type_hints if use_type_hints is not None else True

    arg_to_dependency = _build_arg_to_dependency(arguments, dependencies)
    type_hints = _build_type_hints(arguments, use_type_hints)
    dependency_names = _build_dependency_names(arguments, use_names)

    resolved_dependencies = [
        arg_to_dependency.get(
            arg.name,
            type_hints.get(arg.name,
                           arg.name if arg.name in dependency_names else None)
        )
        for arg in arguments
    ]

    return InjectionBlueprint(tuple([
        Injection(arg_name=arg.name,
                  required=not arg.has_default,
                  dependency=dependency)
        for arg, dependency in zip(arguments, resolved_dependencies)
    ]))


def _build_arg_to_dependency(arguments: Arguments,
                             dependencies: DEPENDENCIES_TYPE = None
                             ) -> Dict[str, Any]:
    if dependencies is None:
        arg_to_dependency = {}  # type: Mapping
    elif isinstance(dependencies, str):
        arg_to_dependency = {arg.name: dependencies.format(arg_name=arg.name)
                             for arg in arguments.without_self}
    elif callable(dependencies):
        arg_to_dependency = {arg.name: dependencies(arg.name)
                             for arg in arguments.without_self}
    elif isinstance(dependencies, c_abc.Mapping):
        _check_valid_arg_names(dependencies.keys(), arguments)
        arg_to_dependency = dependencies
    elif isinstance(dependencies, c_abc.Iterable):
        # convert to Tuple in case we cannot iterate more than once.
        dependencies = tuple(dependencies)
        if len(arguments.without_self) < len(dependencies):
            raise ValueError(("More dependencies ({}) were provided "
                              "than arguments ({})").format(dependencies, arguments))
        arg_to_dependency = {arg.name: dependency
                             for arg, dependency
                             in zip(arguments.without_self, dependencies)}
    else:
        raise TypeError('Only a mapping or a iterable is supported for '
                        'arg_map, not {!r}'.format(type(dependencies)))

    # Remove any None as they would hide type_hints and use_names.
    return {
        k: v
        for k, v in arg_to_dependency.items()
        if v is not None
    }


def _build_type_hints(arguments: Arguments,
                      use_type_hints: Union[bool, Iterable[str]]) -> Dict[str, Any]:
    if use_type_hints is True:
        type_hints = {arg.name: arg.type_hint for arg in arguments.without_self}
    elif use_type_hints is False:
        return {}
    elif isinstance(use_type_hints, c_abc.Iterable):
        # convert to Tuple in case we cannot iterate more than once.
        use_type_hints = tuple(use_type_hints)
        _check_valid_arg_names(use_type_hints, arguments)

        type_hints = {name: arguments[name].type_hint for name in use_type_hints}

    else:
        raise TypeError('Only an iterable or a boolean is supported for '
                        'use_type_hints, not {!r}'.format(type(use_type_hints)))

    # Any object from builtins or typing do not carry any useful information
    # and thus must not be used as dependency IDs. So they might as well be
    # skipped entirely. Moreover they hide use_names.
    return {
        arg_name: type_hint
        for arg_name, type_hint in type_hints.items()
        if getattr(type_hint, '__module__', '') != 'typing'
           and type_hint not in _BUILTINS_TYPES  # noqa
           and type_hint is not None
    }


def _build_dependency_names(arguments: Arguments,
                            use_names: Union[bool, Iterable[str]]) -> Set[str]:
    if use_names is False:
        return set()
    elif use_names is True:
        return {arg.name for arg in arguments.without_self}
    elif isinstance(use_names, c_abc.Iterable):
        # convert to Tuple in case we cannot iterate more than once.
        use_names = tuple(use_names)
        _check_valid_arg_names(use_names, arguments)
        return set(use_names)
    else:
        raise TypeError('Only an iterable or a boolean is supported for '
                        'use_names, not {!r}'.format(type(use_names)))


def _check_valid_arg_names(names: Iterable[str], arguments: Arguments):
    for name in names:
        if not isinstance(name, str):
            raise TypeError("Expected argument name (string), "
                            "got {!r}".format(name))
        if name not in arguments:
            raise ValueError("Unknown argument {!r}".format(name))
        if arguments.has_self and name == arguments[0].name:
            raise ValueError("Cannot inject first argument "
                             "({}) of a method".format(arguments[0]))



import collections.abc as c_abc
import inspect
from typing import Any, Callable, cast, Iterable, overload, TypeVar, Union

from .wire import wire
from .._internal.default_container import get_default_container
from ..core import DEPENDENCIES_TYPE, DependencyContainer, inject
from ..providers.factory import FactoryProvider
from ..providers.tag import Tag, TagProvider

C = TypeVar('C', bound=type)


@overload
def register(class_: C,  # noqa: E704  # pragma: no cover
             *,
             singleton: bool = True,
             factory: Union[Callable, str] = None,
             factory_dependency: Any = None,
             auto_wire: Union[bool, Iterable[str]] = None,
             dependencies: DEPENDENCIES_TYPE = None,
             use_names: Union[bool, Iterable[str]] = None,
             use_type_hints: Union[bool, Iterable[str]] = None,
             wire_super: Union[bool, Iterable[str]] = None,
             tags: Iterable[Union[str, Tag]] = None,
             container: DependencyContainer = None
             ) -> C: ...


@overload
def register(*,  # noqa: E704  # pragma: no cover
             singleton: bool = True,
             factory: Union[Callable, str] = None,
             factory_dependency: Any = None,
             auto_wire: Union[bool, Iterable[str]] = None,
             dependencies: DEPENDENCIES_TYPE = None,
             use_names: Union[bool, Iterable[str]] = None,
             use_type_hints: Union[bool, Iterable[str]] = None,
             wire_super: Union[bool, Iterable[str]] = None,
             tags: Iterable[Union[str, Tag]] = None,
             container: DependencyContainer = None
             ) -> Callable[[C], C]: ...


def register(class_=None,
             *,
             singleton: bool = True,
             factory: Union[Callable, str] = None,
             factory_dependency: Any = None,
             auto_wire: Union[bool, Iterable[str]] = None,
             dependencies: DEPENDENCIES_TYPE = None,
             use_names: Union[bool, Iterable[str]] = None,
             use_type_hints: Union[bool, Iterable[str]] = None,
             wire_super: Union[bool, Iterable[str]] = None,
             tags: Iterable[Union[str, Tag]] = None,
             container: DependencyContainer = None):
    """Register a dependency by its class.

    Args:
        class_: Class to register as a dependency. It will be instantiated
            only when requested.
        singleton: If True, the class will be instantiated only once,
            further will receive the same instance.
        factory: Callable to be used when building the class, this allows to
            re-use the same factory for subclasses for example. The dependency
            is given as first argument. If a string is specified, it is
            interpreted as the name of the method which has to be used to build
            the class. The class is given as first argument for static methods
            but not for class methods. Cannot be used together with
            :code:`factory_dependency`.
        factory_dependency: If specified Antidote will retrieve the factory in
            the container with it. The class is given as first argument. Cannot
            be used together with :code:`factory`.
        auto_wire: Injects automatically the dependencies of the methods
            specified, or only of :code:`__init__()` if True.
        dependencies: Can be either a mapping of arguments name to their
            dependency, an iterable of dependencies or a function which returns
            the dependency given the arguments name. If an iterable is specified,
            the position of the arguments is used to determine their respective
            dependency. An argument may be skipped by using :code:`None` as a
            placeholder. The first argument is always ignored for methods (self)
            and class methods (cls).Type hints are overridden. Defaults to :code:`None`.
        use_names: Whether or not the arguments' name should be used as their
            respective dependency. An iterable of argument names may also be
            supplied to restrict this to those. Defaults to :code:`False`.
        use_type_hints: Whether or not the type hints (annotations) should be
            used as the arguments dependency. An iterable of argument names may
            also be specified to restrict this to those. Any type hints from
            the builtins (str, int...) or the typing (:py:class:`~typing.Optional`,
            ...) are ignored. Defaults to :code:`True`.
        wire_super: If a method from a super-class needs to be wired, specify
            either a list of method names or :code:`True` to enable it for
            all methods. Defaults to :code:`False`, only methods defined in the
            class itself can be wired.
        tags: Iterable of tag to be applied. Those must be either strings
            (the tag name) or :py:class:`~.providers.tag.Tag`. All
            dependencies with a specific tag can then be retrieved with
            a :py:class:`~.providers.tag.Tagged`.
        container: :py:class:`~.core.container.DependencyContainer` to which the
            dependency should be attached. Defaults to the global container,
            :code:`antidote.world`.

    Returns:
        The class or the class decorator.

    """
    if factory is not None and factory_dependency is not None:
        raise ValueError("factory and factory_dependency cannot be used together.")

    if not (factory is None or isinstance(factory, str) or inspect.isfunction(factory)):
        raise TypeError("factory must be either None, a method name or a function "
                        "not {!r}".format(type(factory)))

    container = container or get_default_container()
    auto_wire = auto_wire if auto_wire is not None else True
    methods = ()  # type: Iterable[str]
    wire_raise_on_missing = True

    if isinstance(auto_wire, bool):  # for Mypy
        if auto_wire:
            if isinstance(factory, str):
                methods = (factory,)
                if wire_super is None:
                    wire_super = (factory,)
            elif factory is None:
                wire_raise_on_missing = False
                methods = ('__init__',)
    else:
        methods = auto_wire

    def register_service(cls):
        nonlocal factory

        if not inspect.isclass(cls):
            raise TypeError("Expected a class, got {!r}".format(cls))

        takes_dependency = False
        if isinstance(factory, str):
            static_factory = inspect.getattr_static(cls, factory)
            if not isinstance(static_factory, (staticmethod, classmethod)):
                raise TypeError("Only class methods and static methods "
                                "are supported as factories. Not "
                                "{!r}".format(static_factory))
            if isinstance(static_factory, staticmethod):
                takes_dependency = True

        if auto_wire:
            cls = wire(cls,
                       methods=methods,
                       wire_super=wire_super,
                       dependencies=dependencies,
                       use_names=use_names,
                       use_type_hints=use_type_hints,
                       container=container,
                       raise_on_missing=wire_raise_on_missing)

        if isinstance(factory, str):
            # Retrieve injected class/static method
            factory = cast(Callable, getattr(cls, factory))
        elif inspect.isfunction(factory):
            takes_dependency = True
            if not isinstance(dependencies, c_abc.Mapping) \
                    and isinstance(dependencies, c_abc.Iterable):
                # takes dependency as first argument
                factory_dependencies = (None,) + tuple(dependencies)
            else:
                factory_dependencies = dependencies

            if auto_wire:
                factory = inject(factory,
                                 dependencies=factory_dependencies,
                                 use_names=use_names,
                                 use_type_hints=use_type_hints,
                                 container=container)

        factory_provider = cast(FactoryProvider, container.providers[FactoryProvider])
        if factory is not None:
            factory_provider.register_factory(
                dependency=cls,
                factory=factory,
                singleton=singleton,
                takes_dependency=takes_dependency)
        elif factory_dependency is not None:
            factory_provider.register_providable_factory(
                dependency=cls,
                factory_dependency=factory_dependency,
                singleton=singleton,
                takes_dependency=True)
        else:
            factory_provider.register_class(cls, singleton=singleton)

        if tags is not None:
            tag_provider = cast(TagProvider, container.providers[TagProvider])
            tag_provider.register(cls, tags)

        return cls

    return class_ and register_service(class_) or register_service



# cython: language_level=3
# cython: boundscheck=False, wraparound=False, annotation_typing=False
from typing import Callable, Dict, Tuple, Union

# @formatter:off
from cpython.object cimport PyObject, PyObject_Call, PyObject_GetAttr

from antidote.core.container cimport DependencyInstance, DependencyProvider
# @formatter:on


cdef class LazyCall:
    """
    Dependency which is the result of the call of the given function with the
    given arguments.

    .. doctest::

        >>> from antidote import LazyCall, world
        >>> def f(x, y):
        ...     print("Computing {} + {}".format(x, y))
        ...     return x + y
        >>> A = LazyCall(f)(2, y=3)
        >>> world.get(A)
        Computing 2 + 3
        5
    """
    def __init__(self, func: Callable, singleton: bool = True):
        """
        Args:
            func: Function to lazily call, any arguments given by calling
                to the instance of :py:class:`~.LazyCall` will be passed on.
            singleton: Whether or not this is a singleton or not.
        """
        self._singleton = singleton
        self._func = func
        self._args = ()  # type: Tuple
        self._kwargs = {}  # type: Dict

    def __call__(self, *args, **kwargs):
        """
        All argument are passed on to the lazily called function.
        """
        self._args = args
        self._kwargs = kwargs
        return self

cdef class LazyMethodCall:
    """
    Similar to :py:class:`~.LazyCall` but adapted to methods within a class
    definition. The class has to be a registered service, as the class
    instantiation itself is also lazy.

    .. doctest::

        >>> from antidote import LazyMethodCall, register, world
        >>> @register
        ... class Constants:
        ...     def get(self, x: str):
        ...         return len(x)
        ...     A = LazyMethodCall(get)('test')
        >>> Constants.A
        LazyMethodCallDependency(...)
        >>> world.get(Constants.A)
        4
        >>> Constants().A
        4

    :py:class:`~.LazyMethodCall` has two different behaviors:

    - if retrieved as a class attribute it returns a dependency which identifies
      the result for Antidote.
    - if retrieved as a instance attribute it returns the result for this
      instance. This makes testing a lot easier as it does not require Antidote.

    Check out :py:class:`~.helpers.conf.LazyConstantsMeta` for simple way
    to declare multiple constants.
    """
    def __init__(self, method: Union[Callable, str], singleton: bool = True):
        self._singleton = singleton
        # Retrieve the name of the method, as injection can be done after the class
        # creation which is typically the case with @register.
        self._method_name = method if isinstance(method, str) else method.__name__
        self._args = ()  # type: Tuple
        self._kwargs = {}  # type: Dict
        self._key = None

    def __call__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs
        return self

    def __get__(self, instance, owner):
        if instance is None:
            if self._singleton:
                if self._key is None:
                    self._key = "{}_dependency".format(self._get_attribute_name(owner))
                    setattr(owner, self._key, LazyMethodCallDependency(self, owner))
                return getattr(owner, self._key)
            return LazyMethodCallDependency(self, owner)
        return self._call(instance)

    cdef object _call(self, object instance):
        cdef:
            object method

        method = PyObject_GetAttr(
            instance,
            self._method_name
        )
        if <PyObject*> method == NULL:
            raise RuntimeError("{} does not have a method {}".format(
                instance,
                self._method_name
            ))

        return PyObject_Call(method, self._args, self._kwargs)

    def _get_attribute_name(self, owner):
        for k, v in owner.__dict__.items():
            if v is self:
                return k

cdef class LazyMethodCallDependency:
    cdef:
        LazyMethodCall lazy_method_call
        object owner

    def __cinit__(self, lazy_method_call, owner):
        self.lazy_method_call = lazy_method_call
        self.owner = owner

cdef class LazyCallProvider(DependencyProvider):
    bound_dependency_types = (LazyMethodCallDependency, LazyCall)

    cpdef DependencyInstance provide(self, object dependency):
        cdef:
            LazyCall lazy_call
            LazyMethodCallDependency lazy_method_dependency
            DependencyInstance dependency_instance

        if isinstance(dependency, LazyMethodCallDependency):
            lazy_method_dependency = <LazyMethodCallDependency> dependency
            return DependencyInstance.__new__(
                DependencyInstance,
                lazy_method_dependency.lazy_method_call._call(
                    self._container.get(lazy_method_dependency.owner)
                ),
                lazy_method_dependency.lazy_method_call._singleton
            )
        elif isinstance(dependency, LazyCall):
            lazy_call = <LazyCall> dependency
            return DependencyInstance.__new__(
                DependencyInstance,
                PyObject_Call(lazy_call._func, lazy_call._args, lazy_call._kwargs),
                lazy_call._singleton
            )



import random
import threading
import time

import pytest

from antidote import Tagged, factory, new_container
from antidote.core import DependencyContainer
from antidote.providers.tag import TaggedDependencies


class Service:
    pass


class AnotherService:
    pass


def make_delayed_factory(service, a=0.01, b=0.01):
    def f() -> service:
        time.sleep(a + b * random.random())
        return service()

    return f


@pytest.fixture()
def container():
    c = new_container()
    c.update_singletons({Service: Service(), 'parameter': object()})

    return c


def multi_thread_do(target, n_threads=10):
    threads = [threading.Thread(target=target)
               for _ in range(n_threads)]

    for thread in threads:
        thread.start()

    for thread in threads:
        thread.join()


def test_container_instantiation_safety(container: DependencyContainer):
    n_threads = 10

    factory(make_delayed_factory(Service),
            singleton=True,
            container=container)
    factory(make_delayed_factory(AnotherService),
            singleton=False,
            container=container)

    singleton_got = []
    non_singleton_got = []

    def worker():
        singleton_got.append(container.get(Service))
        non_singleton_got.append(container.get(AnotherService))

    multi_thread_do(worker, n_threads)

    assert 1 == len(set(singleton_got))
    assert n_threads == len(set(non_singleton_got))


def test_tagged_dependencies_instantiation_safety(container: DependencyContainer):
    n_dependencies = 40

    for i in range(n_dependencies):
        factory(make_delayed_factory(type('Service{}'.format(i), (object,), {})),
                singleton=False,
                tags=['test'],
                container=container)

    tagged = container.get(Tagged('test'))  # type: TaggedDependencies
    dependencies = []

    def worker():
        for i, dep in enumerate(tagged.instances()):
            dependencies.append((i, dep))

    multi_thread_do(worker)

    assert n_dependencies == len(set(dependencies))
    assert set(dependencies) == set(enumerate(tagged.instances()))



import os
import pathlib

from setuptools import Extension, find_packages, setup

here = pathlib.Path(os.path.dirname(os.path.abspath(__file__)))

with open(str(here / 'README.rst'), 'r') as f:
    readme = f.read()


def generate_extensions():
    extensions = []
    for root, _, filenames in os.walk('src'):
        for filename in filenames:
            if filename.endswith('.pyx'):
                path = os.path.join(root, filename)
                module = path[4:].replace('/', '.').rsplit('.', 1)[0]
                extensions.append(Extension(module,
                                            [path],
                                            language='c++'))
    return extensions


ext_modules = []
requires = []
setup_requires = ['setuptools_scm']

try:
    from Cython.Build import cythonize
except ImportError:
    pass
else:
    ext_modules = cythonize(generate_extensions())
    requires.append('fastrlock>=0.4,<0.5')
    setup_requires.append('fastrlock>=0.4,<0.5')

setup(
    name='antidote',
    use_scm_version=True,
    setup_requires=setup_requires,
    description='Transparent dependency injection.',
    long_description=readme,
    author='Benjamin Rabier',
    url='https://github.com/Finistere/antidote',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_dirs=["src"],
    ext_modules=ext_modules,
    install_requires=requires,
    extras_require={
        ":python_version<'3.5'": ["typing"],
    },
    license='MIT',
    classifiers=(
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
    ),
    keywords='dependency injection',
    zip_safe=False,
)



from typing import Any

import pytest

from antidote.core import DependencyContainer, DependencyInstance, DependencyProvider
from antidote.exceptions import (DependencyCycleError, DependencyInstantiationError,
                                 DependencyNotFoundError)
from .utils import DummyFactoryProvider, DummyProvider


class Service:
    def __init__(self, *args):
        pass


class AnotherService:
    def __init__(self, *args):
        pass


class YetAnotherService:
    def __init__(self, *args):
        pass


class ServiceWithNonMetDependency:
    def __init__(self, dependency):
        pass


@pytest.fixture()
def container():
    return DependencyContainer()


def test_dependency_repr():
    o = object()
    d = DependencyInstance(o, False)

    assert repr(False) in repr(d)
    assert repr(o) in repr(d)


def test_setitem(container: DependencyContainer):
    s = object()
    container.update_singletons({'service': s})

    assert s is container.get('service')
    assert repr(s) in repr(container)


def test_update(container: DependencyContainer):
    container.update_singletons({Service: Service()})

    another_service = AnotherService()
    x = object()

    container.update_singletons({
        Service: another_service,
        'x': x
    })

    assert another_service is container.get(Service)
    assert x is container.get('x')


def test_register_provider(container: DependencyContainer):
    provider = DummyProvider()
    container.register_provider(provider)

    # Can't check directly if the core is the same, as a proxy is used
    assert provider is container.providers[DummyProvider]


def test_getitem(container: DependencyContainer):
    container.register_provider(DummyFactoryProvider({
        Service: lambda: Service(),
        ServiceWithNonMetDependency: lambda: ServiceWithNonMetDependency(),
    }))
    container.register_provider(DummyProvider({'name': 'Antidote'}))

    with pytest.raises(DependencyNotFoundError):
        container.get(object)

    with pytest.raises(DependencyInstantiationError):
        container.get(ServiceWithNonMetDependency)

    assert isinstance(container.get(Service), Service)
    assert isinstance(container.provide(Service), DependencyInstance)
    assert 'Antidote' == container.get('name')
    assert 'Antidote' == container.provide('name').instance


def test_singleton(container: DependencyContainer):
    container.register_provider(DummyFactoryProvider({
        Service: lambda: Service(),
        AnotherService: lambda: AnotherService(),
    }))

    service = container.get(Service)
    assert service is container.get(Service)

    container.providers[DummyFactoryProvider].singleton = False
    another_service = container.get(AnotherService)
    assert another_service is not container.get(AnotherService)

    singletons = {k: v.instance for k, v in container.singletons.items()}
    assert {Service: service, DependencyContainer: container} == singletons


def test_dependency_cycle_error(container: DependencyContainer):
    container.register_provider(DummyFactoryProvider({
        Service: lambda: Service(container.get(AnotherService)),
        AnotherService: lambda: AnotherService(container.get(YetAnotherService)),
        YetAnotherService: lambda: YetAnotherService(container.get(Service)),
    }))

    with pytest.raises(DependencyCycleError):
        container.get(Service)

    with pytest.raises(DependencyCycleError):
        container.get(AnotherService)

    with pytest.raises(DependencyCycleError):
        container.get(YetAnotherService)


def test_repr_str(container: DependencyContainer):
    container.register_provider(DummyProvider({'name': 'Antidote'}))
    container.update_singletons({'test': 1})

    assert 'test' in repr(container)
    assert repr(container.providers[DummyProvider]) in repr(container)
    assert str(container.providers[DummyProvider]) in str(container)


def test_invalid_provider(container: DependencyContainer):
    with pytest.raises(TypeError):
        container.register_provider(object())


def test_bound_dependency_types():
    class CustomDependency:
        pass

    class DummyProvider1(DependencyProvider):
        bound_dependency_types = (CustomDependency,)

        def provide(self, dependency: Any) -> DependencyInstance:
            return DependencyInstance(self)

    class DummyProvider2(DependencyProvider):
        def provide(self, dependency: Any) -> DependencyInstance:
            raise Exception()

    container = DependencyContainer()
    container.register_provider(DummyProvider2(container))
    container.register_provider(DummyProvider1(container))
    assert isinstance(container.get(CustomDependency()), DummyProvider1)

    container = DependencyContainer()
    container.register_provider(DummyProvider1(container))
    container.register_provider(DummyProvider2(container))
    assert isinstance(container.get(CustomDependency()), DummyProvider1)


def test_bound_dependency_types_conflict():
    class CustomDependency:
        pass

    class DummyProvider1(DependencyProvider):
        bound_dependency_types = (CustomDependency,)

        def provide(self, dependency: Any) -> DependencyInstance:
            return DependencyInstance(self)

    class DummyProvider2(DependencyProvider):
        bound_dependency_types = (CustomDependency,)

        def provide(self, dependency: Any) -> DependencyInstance:
            return DependencyInstance(self)

    container = DependencyContainer()
    container.register_provider(DummyProvider1(container))

    with pytest.raises(RuntimeError):
        container.register_provider(DummyProvider2(container))



import pytest

from antidote import new_container, provider
from antidote.core import DependencyInstance, DependencyProvider
from antidote.providers import (IndirectProvider, LazyCallProvider, FactoryProvider,
                                TagProvider)


@pytest.fixture()
def container():
    return new_container()


def test_simple(container):
    container.update_singletons({'service': object()})

    @provider(container=container, wire_super=True)
    class DummyProvider(DependencyProvider):
        def provide(self, dependency):
            if dependency == 'test':
                return DependencyInstance(dependency)

    assert isinstance(container.providers[DummyProvider], DummyProvider)
    assert 'test' == container.get('test')


@pytest.mark.parametrize('cls', [1, type('MissingCall', tuple(), {})])
def test_invalid_provider(cls):
    with pytest.raises(TypeError):
        provider(cls)


def test_providers(container):
    assert 4 == len(container.providers)
    assert FactoryProvider in container.providers
    assert TagProvider in container.providers
    assert LazyCallProvider in container.providers
    assert IndirectProvider in container.providers

    @provider(container=container, wire_super=True)
    class DummyProvider(DependencyProvider):
        def provide(self, dependency):
            return DependencyInstance(1)

    assert DummyProvider in container.providers



from typing import Callable, Iterable, overload, TypeVar, Union, Type

from .wire import wire
from .._internal.default_container import get_default_container
from ..core import DEPENDENCIES_TYPE, DependencyContainer, DependencyProvider

P = TypeVar('P', bound=Type[DependencyProvider])


@overload
def provider(class_: P,  # noqa: E704  # pragma: no cover
             *,
             auto_wire: Union[bool, Iterable[str]] = True,
             dependencies: DEPENDENCIES_TYPE = None,
             use_names: Union[bool, Iterable[str]] = None,
             use_type_hints: Union[bool, Iterable[str]] = None,
             wire_super: Union[bool, Iterable[str]] = None,
             container: DependencyContainer = None
             ) -> P: ...


@overload
def provider(*,  # noqa: E704  # pragma: no cover
             auto_wire: Union[bool, Iterable[str]] = True,
             dependencies: DEPENDENCIES_TYPE = None,
             use_names: Union[bool, Iterable[str]] = None,
             use_type_hints: Union[bool, Iterable[str]] = None,
             wire_super: Union[bool, Iterable[str]] = None,
             container: DependencyContainer = None
             ) -> Callable[[P], P]: ...


def provider(class_: Type[DependencyProvider] = None,
             *,
             auto_wire: Union[bool, Iterable[str]] = True,
             dependencies: DEPENDENCIES_TYPE = None,
             use_names: Union[bool, Iterable[str]] = None,
             use_type_hints: Union[bool, Iterable[str]] = None,
             wire_super: Union[bool, Iterable[str]] = None,
             container: DependencyContainer = None):
    """Register a providers by its class.

    Args:
        class_: class to register as a provider. The class must have a
            :code:`__antidote_provide()` method accepting as first argument
            the dependency. Variable keyword and positional arguments must be
            accepted as they may be provided.
        auto_wire: If True, the dependencies of :code:`__init__()` are
            injected. An iterable of method names which require dependency
            injection may also be specified.
        dependencies: Can be either a mapping of arguments name to their
            dependency, an iterable of dependencies or a function which returns
            the dependency given the arguments name. If an iterable is specified,
            the position of the arguments is used to determine their respective
            dependency. An argument may be skipped by using :code:`None` as a
            placeholder. The first argument is always ignored for methods (self)
            and class methods (cls).Type hints are overridden. Defaults to :code:`None`.
        use_names: Whether or not the arguments' name should be used as their
            respective dependency. An iterable of argument names may also be
            supplied to restrict this to those. Defaults to :code:`False`.
        use_type_hints: Whether or not the type hints (annotations) should be
            used as the arguments dependency. An iterable of argument names may
            also be specified to restrict this to those. Any type hints from
            the builtins (str, int...) or the typing (:py:class:`~typing.Optional`,
            ...) are ignored. Defaults to :code:`True`.
        wire_super: If a method from a super-class needs to be wired, specify
            either a list of method names or :code:`True` to enable it for
            all methods. Defaults to :code:`False`, only methods defined in the
            class itself can be wired.
        container: :py:class:`~.core.container.DependencyContainer` to which the
            dependency should be attached. Defaults to the global container,
            :code:`antidote.world`.

    Returns:
        the providers's class or the class decorator.
    """
    container = container or get_default_container()

    def register_provider(cls):
        if not issubclass(cls, DependencyProvider):
            raise TypeError("A provider must be subclass of Provider.")

        if auto_wire:
            cls = wire(cls,
                       methods=(('__init__',)
                                if auto_wire is True else
                                auto_wire),
                       dependencies=dependencies,
                       wire_super=wire_super,
                       use_names=use_names,
                       use_type_hints=use_type_hints,
                       container=container,
                       raise_on_missing=auto_wire is not True)

        container.register_provider(cls(container=container))

        return cls

    return class_ and register_provider(class_) or register_provider



import inspect
from typing import (Callable, cast, get_type_hints, Iterable, overload, TypeVar, Union)

from .register import register
from .wire import wire
from .._internal.default_container import get_default_container
from ..core import DEPENDENCIES_TYPE, DependencyContainer, inject
from ..exceptions import DuplicateDependencyError
from ..providers.factory import FactoryProvider
from ..providers.tag import Tag, TagProvider

F = TypeVar('F', Callable, type)


@overload
def factory(func: F,  # noqa: E704  # pragma: no cover
            *,
            auto_wire: Union[bool, Iterable[str]] = True,
            singleton: bool = True,
            dependencies: DEPENDENCIES_TYPE = None,
            use_names: Union[bool, Iterable[str]] = None,
            use_type_hints: Union[bool, Iterable[str]] = None,
            wire_super: Union[bool, Iterable[str]] = None,
            tags: Iterable[Union[str, Tag]] = None,
            container: DependencyContainer = None
            ) -> F: ...


@overload
def factory(*,  # noqa: E704  # pragma: no cover
            auto_wire: Union[bool, Iterable[str]] = True,
            singleton: bool = True,
            dependencies: DEPENDENCIES_TYPE = None,
            use_names: Union[bool, Iterable[str]] = None,
            use_type_hints: Union[bool, Iterable[str]] = None,
            wire_super: Union[bool, Iterable[str]] = None,
            tags: Iterable[Union[str, Tag]] = None,
            container: DependencyContainer = None
            ) -> Callable[[F], F]: ...


def factory(func: Union[Callable, type] = None,
            *,
            auto_wire: Union[bool, Iterable[str]] = True,
            singleton: bool = True,
            dependencies: DEPENDENCIES_TYPE = None,
            use_names: Union[bool, Iterable[str]] = None,
            use_type_hints: Union[bool, Iterable[str]] = None,
            wire_super: Union[bool, Iterable[str]] = None,
            tags: Iterable[Union[str, Tag]] = None,
            container: DependencyContainer = None
            ):
    """Register a dependency providers, a factory to build the dependency.

    Args:
        func: Callable which builds the dependency.
        singleton: If True, `func` will only be called once. If not it is
            called at each injection.
        auto_wire: If :code:`func` is a function, its dependencies are
            injected if True. Should :code:`func` be a class with
            :py:func:`__call__`, dependencies of :code:`__init__()` and
            :code:`__call__()` will be injected if True. One may also
            provide an iterable of method names requiring dependency
            injection.
        dependencies: Can be either a mapping of arguments name to their
            dependency, an iterable of dependencies or a function which returns
            the dependency given the arguments name. If an iterable is specified,
            the position of the arguments is used to determine their respective
            dependency. An argument may be skipped by using :code:`None` as a
            placeholder. The first argument is always ignored for methods (self)
            and class methods (cls).Type hints are overridden. Defaults to :code:`None`.
        use_names: Whether or not the arguments' name should be used as their
            respective dependency. An iterable of argument names may also be
            supplied to restrict this to those. Defaults to :code:`False`.
        use_type_hints: Whether or not the type hints (annotations) should be
            used as the arguments dependency. An iterable of argument names may
            also be specified to restrict this to those. Any type hints from
            the builtins (str, int...) or the typing (:py:class:`~typing.Optional`,
            ...) are ignored. Defaults to :code:`True`.
        wire_super: If a method from a super-class needs to be wired, specify
            either a list of method names or :code:`True` to enable it for
            all methods. Defaults to :code:`False`, only methods defined in the
            class itself can be wired.
        tags: Iterable of tag to be applied. Those must be either strings
            (the tag name) or :py:class:`~.providers.tag.Tag`. All
            dependencies with a specific tag can then be retrieved with
            a :py:class:`~.providers.tag.Tagged`.
        container: :py:class:`~.core.container.DependencyContainer` to which the
            dependency should be attached. Defaults to the global container,
            :code:`antidote.world`.

    Returns:
        object: The dependency_provider

    """
    container = container or get_default_container()

    def register_factory(obj):
        factory_provider = cast(FactoryProvider,
                                container.providers[FactoryProvider])

        if inspect.isclass(obj):
            if '__call__' not in dir(obj):
                raise TypeError("The class must implement __call__()")

            wire_raise_on_missing = True
            if auto_wire is None or isinstance(auto_wire, bool):
                if auto_wire is False:
                    methods = ()  # type: Iterable[str]
                else:
                    methods = ('__call__', '__init__')
                    wire_raise_on_missing = False
            else:
                methods = auto_wire

            if methods:
                obj = wire(obj,
                           methods=methods,
                           wire_super=wire_super,
                           raise_on_missing=wire_raise_on_missing,
                           dependencies=dependencies,
                           use_names=use_names,
                           use_type_hints=use_type_hints,
                           container=container)

            obj = register(obj, auto_wire=False, singleton=True, container=container)
            dependency = get_type_hints(obj.__call__).get('return')
            if dependency is None:
                raise ValueError("The return annotation is necessary on __call__."
                                 "It is used a the dependency.")
            factory_provider.register_providable_factory(
                dependency=dependency,
                singleton=singleton,
                takes_dependency=False,
                factory_dependency=obj
            )
        elif callable(obj):
            if auto_wire:
                obj = inject(obj,
                             dependencies=dependencies,
                             use_names=use_names,
                             use_type_hints=use_type_hints,
                             container=container)

            dependency = get_type_hints(obj).get('return')
            if dependency is None:
                raise ValueError("A return annotation is necessary."
                                 "It is used a the dependency.")
            factory_provider.register_factory(factory=obj,
                                              singleton=singleton,
                                              dependency=dependency,
                                              takes_dependency=False)
        else:
            raise TypeError("Must be either a function "
                            "or a class implementing __call__(), "
                            "not {!r}".format(type(obj)))

        if tags is not None:
            tag_provider = cast(TagProvider, container.providers[TagProvider])
            tag_provider.register(dependency=dependency,
                                  tags=tags)

        return obj

    return func and register_factory(func) or register_factory



# cython: language_level=3
# cython: boundscheck=False, wraparound=False

cdef class DependencyStack:
    cdef:
        list _stack
        set _seen

    cdef bint push(self, object dependency)
    cdef pop(self)



import itertools
from inspect import getattr_static

import pytest
from pretend import raiser

from antidote._internal.argspec import Argument, Arguments


def f(a: str, b, c: int = 1):
    pass


def g(a: list, *args):
    pass


def h(b=None, **kwargs):
    pass


def k():
    pass


class Dummy:
    def f(self, a: str, b=1, *args, **kwargs):
        pass

    @classmethod
    def g(cls, a):
        pass

    @staticmethod
    def h():
        pass


dummy = Dummy()


def lazy(dummy: 'Dummy'):
    pass


@pytest.mark.parametrize(
    'func,expected',
    [
        pytest.param(
            f,
            Arguments(
                arguments=tuple([
                    Argument('a', False, str),
                    Argument('b', False, None),
                    Argument('c', True, int),
                ]),
                has_var_positional=False,
                has_var_keyword=False,
                has_self=False
            ),
            id='f'
        ),
        pytest.param(
            g,
            Arguments(
                arguments=tuple([
                    Argument('a', False, list),
                ]),
                has_var_positional=True,
                has_var_keyword=False,
                has_self=False
            ),
            id='g'
        ),
        pytest.param(
            h,
            Arguments(
                arguments=tuple([
                    Argument('b', True, None),
                ]),
                has_var_positional=False,
                has_var_keyword=True,
                has_self=False
            ),
            id='h'
        ),
        pytest.param(
            k,
            Arguments(
                arguments=tuple([]),
                has_var_positional=False,
                has_var_keyword=False,
                has_self=False
            ),
            id='k'
        ),
        pytest.param(
            lazy,
            Arguments(
                arguments=tuple([
                    Argument('dummy', False, Dummy),
                ]),
                has_var_positional=False,
                has_var_keyword=False,
                has_self=False
            ),
            id='lazy'
        ),
        pytest.param(
            Dummy.f,
            Arguments(
                arguments=tuple([
                    Argument('self', False, None),
                    Argument('a', False, str),
                    Argument('b', True, None),
                ]),
                has_var_positional=True,
                has_var_keyword=True,
                has_self=True
            ),
            id='cls.f'
        ),
        pytest.param(
            Dummy.g,
            Arguments(
                arguments=tuple([
                    Argument('a', False, None),
                ]),
                has_var_positional=False,
                has_var_keyword=False,
                has_self=False
            ),
            id='cls.g'
        ),
        pytest.param(
            getattr_static(Dummy, 'g'),
            Arguments(
                arguments=tuple([
                    Argument('cls', False, None),
                    Argument('a', False, None),
                ]),
                has_var_positional=False,
                has_var_keyword=False,
                has_self=True
            ),
            id='cls.g'
        ),
        pytest.param(
            Dummy.h,
            Arguments(
                arguments=tuple([]),
                has_var_positional=False,
                has_var_keyword=False,
                has_self=False
            ),
            id='cls.h'
        ),
        pytest.param(
            dummy.f,
            Arguments(
                arguments=tuple([
                    Argument('a', False, str),
                    Argument('b', True, None),
                ]),
                has_var_positional=True,
                has_var_keyword=True,
                has_self=False
            ),
            id='instance.f'
        ),
        pytest.param(
            dummy.g,
            Arguments(
                arguments=tuple([
                    Argument('a', False, None),
                ]),
                has_var_positional=False,
                has_var_keyword=False,
                has_self=False
            ),
            id='instance.g'
        ),
        pytest.param(
            dummy.h,
            Arguments(
                arguments=tuple([]),
                has_var_positional=False,
                has_var_keyword=False,
                has_self=False
            ),
            id='instance.h'
        ),
    ]
)
def test_from_callable(func, expected: Arguments):
    result = Arguments.from_callable(func)
    assert isinstance(result, Arguments)
    assert expected.has_var_positional == result.has_var_positional
    assert expected.has_var_keyword == result.has_var_keyword

    for expected_arg, result_arg in zip(expected, result):
        assert expected_arg.name == result_arg.name
        assert expected_arg.has_default == result_arg.has_default
        assert expected_arg.type_hint == result_arg.type_hint

    for expected_arg in expected:
        assert expected_arg.name in result

        result_arg = result[expected_arg.name]
        assert expected_arg.name == result_arg.name
        assert expected_arg.has_default == result_arg.has_default
        assert expected_arg.type_hint == result_arg.type_hint


@pytest.mark.parametrize(
    'descriptor',
    [staticmethod, classmethod]
)
def test_from_methods(descriptor):
    def f(a, b=None):
        pass

    expected = Arguments.from_callable(f)
    result = Arguments.from_callable(descriptor(f))

    for expected_arg, result_arg in zip(expected, result):
        assert expected_arg.name == result_arg.name
        assert expected_arg.has_default == result_arg.has_default
        assert expected_arg.type_hint == result_arg.type_hint


def test_broken_type_hints_cpy353(monkeypatch):
    monkeypatch.setattr('antidote._internal.argspec.get_type_hints', raiser(Exception))
    Arguments.from_callable(k)


args = tuple([
    Argument('x', False, int),
    Argument('y', True, str),
    Argument('z', False, float),
])


def test_getitem():
    arguments = Arguments(
        arguments=args,
        has_var_keyword=False,
        has_var_positional=False,
        has_self=False
    )
    assert arguments['x'] is args[0]
    assert arguments[1] is args[1]

    with pytest.raises(TypeError):
        arguments[2.3]


def test_without_self():
    for has_var_keyword, has_var_positional in itertools.product([True, False],
                                                                 [True, False]):
        arguments = Arguments(
            arguments=args,
            has_var_keyword=has_var_keyword,
            has_var_positional=has_var_positional,
            has_self=True
        )
        assert tuple(arguments.without_self) == args[1:]
        assert arguments.without_self.has_var_keyword == arguments.has_var_keyword
        assert arguments.without_self.has_var_positional == arguments.has_var_positional

    arguments = Arguments(
        arguments=args,
        has_var_keyword=False,
        has_var_positional=False,
        has_self=False
    )
    assert arguments.without_self is arguments


def test_magic_methods_arguments():
    arguments = Arguments(
        arguments=args,
        has_var_keyword=False,
        has_var_positional=False,
        has_self=False
    )
    assert 'x' in arguments
    assert 'unknown' not in arguments
    assert 3 == len(arguments)
    assert args == tuple(arguments)
    assert "x:int" in repr(arguments)
    assert "y:str =" in repr(arguments)
    assert "z:float" in repr(arguments)

    arguments_var_args = Arguments(
        arguments=args,
        has_var_keyword=False,
        has_var_positional=True,
        has_self=False
    )
    assert "x:int" in repr(arguments_var_args)
    assert "y:str =" in repr(arguments_var_args)
    assert "z:float" in repr(arguments_var_args)
    assert "*args" in repr(arguments_var_args)

    arguments_var_kwargs = Arguments(
        arguments=args,
        has_var_keyword=True,
        has_var_positional=False,
        has_self=False
    )
    assert "x:int" in repr(arguments_var_kwargs)
    assert "y:str =" in repr(arguments_var_kwargs)
    assert "z:float" in repr(arguments_var_kwargs)
    assert "**kwargs" in repr(arguments_var_kwargs)



# cython: language_level=3
# cython: boundscheck=False, wraparound=False
from antidote.core.container cimport DependencyInstance, DependencyProvider

cdef class FactoryProvider(DependencyProvider):
    cdef:
        dict _builders

    cpdef DependencyInstance provide(self, object dependency)

cdef class Build:
    cdef:
        readonly object dependency
        readonly dict kwargs
        int _hash



import pytest

from antidote._internal.stack import DependencyStack
from antidote.exceptions import DependencyCycleError


class Service:
    pass


class CustomException(Exception):
    pass


def test_instantiating():
    ds = DependencyStack()

    with ds.instantiating(DependencyStack):
        with ds.instantiating('test'):
            pass

    with pytest.raises(DependencyCycleError):
        with ds.instantiating(DependencyStack):
            with ds.instantiating(DependencyStack):
                pass

    try:
        with ds.instantiating(DependencyStack):
            raise CustomException()
    except CustomException:
        pass

    # DependencyStack should be clean
    with ds.instantiating(DependencyStack):
        pass
